-- There is no copyright on this code

-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
-- associated documentation files (the "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is furnished to do so.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
-- NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

--
-- Global variables
--
local addon, XMacros = ...
XMacros.AbilityManager = {}
XMacros.AbilityManager.name = "XMacros.AbilityManager"
XMacros.AbilityManager.abilityByID = {}
XMacros.AbilityManager.abilityByName = {}
XMacros.AbilityManager.abilityCallbacks = {}
XMacros.AbilityManager.abilityCallbackMap = {}
XMacros.AbilityManager.callbackID = 0
XMacros.AbilityManager.abilityCount = 0
XMacros.AbilityManager.beginType = 1
XMacros.AbilityManager.endType = 2
--
-- Main functions
--

local function AbilityDetail(abilityID)
	if Inspect.Ability.New ~= nil then
		return Inspect.Ability.New.Detail(abilityID)
	else
		return Inspect.Ability.Detail(abilityID)
	end
end

local function AbilityList()
	if Inspect.Ability.New ~= nil then
		return Inspect.Ability.New.List()
	else
		return Inspect.Ability.List()
	end
end

function XMacros.AbilityManager.GetAbilities()
	local abilities = AbilityList()
	if abilities then
		return AbilityDetail(abilities)
	end
end

function XMacros.AbilityManager.AddEvents()
	if Event.Ability.New ~= nil then
		table.insert(Event.Ability.New.Cooldown.Begin, XMacros.AbilityManager.CooldownBeginHook)
		table.insert(Event.Ability.New.Cooldown.End, XMacros.AbilityManager.CooldownEndHook)
	else
		table.insert(Event.Ability.Cooldown.Begin, XMacros.AbilityManager.CooldownBeginHook)
		table.insert(Event.Ability.Cooldown.End, XMacros.AbilityManager.CooldownEndHook)
	end
end

function XMacros.AbilityManager.RemoveEvent(tableName, inHook)
	for i, hook in ipairs(tableName) do
		if hook == inHook then
			table.remove(tableName, i)
			return
		end
	end
end

function XMacros.AbilityManager.RemoveEvents()
	XMacros.AbilityManager.RemoveEvent(Event.Ability.Cooldown.Begin, XMacros.AbilityManager.CooldownBeginHook)
	XMacros.AbilityManager.RemoveEvent(Event.Ability.Cooldown.End, XMacros.AbilityManager.CooldownEndHook)
end

function XMacros.AbilityManager.Clear()
	XMacros.AbilityManager.RemoveEvents()
	XMacros.AbilityManager.abilityCallbacks = {}
	XMacros.AbilityManager.abilityCallbackMap = {}
	XMacros.AbilityManager.abilityCount = 0
end

function XMacros.AbilityManager.MapAbility(abilityID, detail)
	if detail and detail.name then
		XMacros.AbilityManager.abilityByID[abilityID] = detail.name
		XMacros.AbilityManager.abilityByName[detail.name] = abilityID
	end
end

function XMacros.AbilityManager.UnmapAbility(abilityID)
	local name = XMacros.AbilityManager.abilityByID[abilityID]
	if name then
		XMacros.AbilityManager.abilityByName[name] = nil
	end
	
	XMacros.AbilityManager.abilityByID[abilityID] = nil
end

function XMacros.AbilityManager.AddAbilities(abilityIDs)
	if abilityIDs then
		for abilityID, _ in pairs(abilityIDs) do
			if not XMacros.AbilityManager.abilityByID[abilityID] then
				local detail = AbilityDetail(abilityID)
				if detail then
					XMacros.AbilityManager.MapAbility(abilityID, detail)
				end
			end
		end
	end
end

function XMacros.AbilityManager.RemoveAbilities(abilityIDs)
	if abilityIDs then
		for abilityID, _ in pairs(abilityIDs) do
			if XMacros.AbilityManager.abilityByID[abilityID] then
				XMacros.AbilityManager.UnmapAbility(abilityID)
			end
		end
	end
end

function XMacros.AbilityManager.GetCooldown(abilityName)
	if not XMacros.AbilityManager.abilityByName[abilityName] then
		return nil
	end
	
	local abilityID = XMacros.AbilityManager.abilityByName[abilityName]
	if abilityID then
		local details = AbilityDetail(abilityID)
		if details ~= nil then
			return details.currentCooldownBegin
		end
	end
	
	return nil
end

function XMacros.AbilityManager.GetID(abilityName)
	return XMacros.AbilityManager.abilityByName[abilityName]
end

function XMacros.AbilityManager.GetName(abilityID)
	return XMacros.AbilityManager.abilityByID[abilityID]
end

function XMacros.AbilityManager.GetAbilityNames()
	local names = {}
	for name, _ in pairs(XMacros.AbilityManager.abilityByName) do
		table.insert(names, name)
	end
	
	table.sort(names)
	
	return names
end

function XMacros.AbilityManager.AddAbilityBeginCallback(abilityName, callback)
	return XMacros.AbilityManager.InternalAddAbilityCallback(abilityName, callback, XMacros.AbilityManager.beginType)
end

function XMacros.AbilityManager.RemoveAbilityCallback(id)
	local key = XMacros.AbilityManager.abilityCallbackMap[id]
	if key then
		XMacros.AbilityManager.abilityCount = XMacros.AbilityManager.abilityCount - 1
		if XMacros.AbilityManager.abilityCount == 0 then
			XMacros.AbilityManager.Clear()
		else
			XMacros.AbilityManager.abilityCallbacks[key][id] = nil
		end
	end
end

function XMacros.AbilityManager.AddAbilityEndCallback(abilityName, callback)
	return XMacros.AbilityManager.InternalAddAbilityCallback(abilityName, callback, XMacros.AbilityManager.endType)
end

function XMacros.AbilityManager.InternalAddAbilityCallback(abilityName, callback, eventType)
	if not abilityName then
		print("Ignoring ability with no name")
		return -1
	end
	if not abilityName then
		print("Ignoring " .. unitSpec .. " ability with no name")
		return -1
	end
	
	local abilityID = XMacros.AbilityManager.GetID(abilityName)
	if not abilityID then
		return -1
	end
	
	local details = AbilityDetail(abilityID)
	
	if not details or not details.cooldown or details.cooldown == 0 then
		return -1
	end
	
	if XMacros.AbilityManager.abilityCount == 0 then
		XMacros.AbilityManager.AddEvents()
	end
	
	XMacros.AbilityManager.abilityCount = XMacros.AbilityManager.abilityCount + 1
	
	local id = XMacros.AbilityManager.GetCallbackID()
	XMacros.AbilityManager.abilityCallbackMap[id] = abilityName
	
	if not XMacros.AbilityManager.abilityCallbacks[abilityName] then
		XMacros.AbilityManager.abilityCallbacks[abilityName] = {}
	end

	XMacros.AbilityManager.abilityCallbacks[abilityName][id] = {}
	XMacros.AbilityManager.abilityCallbacks[abilityName][id].callback = callback
	XMacros.AbilityManager.abilityCallbacks[abilityName][id].eventType = eventType
	
	XMacros.AbilityManager.Refresh(abilityName)
	return id
end

function XMacros.AbilityManager.GetCallbackID()
	XMacros.AbilityManager.callbackID = XMacros.AbilityManager.callbackID + 1
	return "#" .. XMacros.AbilityManager.callbackID
end

function XMacros.AbilityManager.Refresh(abilityName)
	if not XMacros.AbilityManager.abilityCallbacks[abilityName] then
		return
	end

	local abilityID = XMacros.AbilityManager.GetID(abilityName)
	if not abilityID then
		return
	end	
end


function XMacros.AbilityManager.AbilityFired(abilities)
	local frameTime = nil
	
	if XMacros.AbilityManager.abilityCount > 0 then
		for abilityID, count in pairs(abilities) do
			local abilityName = XMacros.AbilityManager.GetName(abilityID)
			if abilityName and XMacros.AbilityManager.abilityCallbacks[abilityName] then
				local fireEvent = true
				local details = AbilityDetail(abilityID)
				local isStarted = false
				local isEnded = false

				local remaining = 0
				if details and details.currentCooldownRemaining then
					remaining = XMacros.Utils.Round(details.currentCooldownRemaining, 2)
				end
				
				if remaining == 0 then
					isEnded = true
				else
					isStarted = true
				end

				if fireEvent == true then
					if not frameTime then
						frameTime = Inspect.Time.Frame()
					end
					
					for id, callbackTable in pairs(XMacros.AbilityManager.abilityCallbacks[abilityName]) do
						if callbackTable.eventType == XMacros.AbilityManager.beginType and isStarted == true then
							callbackTable.callback(abilityName, remaining, frameTime)
						elseif callbackTable.eventType == XMacros.AbilityManager.endType and isEnded == true then
							callbackTable.callback(abilityName)
						end
					end
				end
			end
		end
	end
end

--
-- Event Handlers
--

function XMacros.AbilityManager.AbilityBegin(abilities)
	XMacros.AbilityManager.AbilityFired(abilities)
end

function XMacros.AbilityManager.AbilityEnd(abilities)
	XMacros.AbilityManager.AbilityFired(abilities)
end

XMacros.AbilityManager.AddHook = { XMacros.AbilityManager.AddAbilities, "XMacros", "AbilityAdd" }
XMacros.AbilityManager.RemoveHook = { XMacros.AbilityManager.RemoveAbilities, "XMacros", "AbilityRemove" }
XMacros.AbilityManager.CooldownBeginHook = { XMacros.AbilityManager.AbilityBegin, "XMacros", "AbilityBegin" }
XMacros.AbilityManager.CooldownEndHook = { XMacros.AbilityManager.AbilityEnd, "XMacros", "AbilityEnd" }

if Event.Ability.New ~= nil then
	table.insert(Event.Ability.New.Add, XMacros.AbilityManager.AddHook)
	table.insert(Event.Ability.New.Remove, XMacros.AbilityManager.RemoveHook)
else
	table.insert(Event.Ability.Add, XMacros.AbilityManager.AddHook)
	table.insert(Event.Ability.Remove, XMacros.AbilityManager.RemoveHook)
end
