-- There is no copyright on this code

-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
-- associated documentation files (the "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is furnished to do so.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
-- NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

--
-- Global variables
--
local addon, XMacros = ...
XMacros.ButtonSets = {}
XMacros.ButtonSets.ShowTypeAlways = 1
XMacros.ButtonSets.ShowTypeInCombat = 2
XMacros.ButtonSets.ShowTypeOutCombat = 3
XMacros.ButtonSets.ShowTypeMouse = 4
XMacros.ButtonSets.ShowTypeInCombatMouse = 5
XMacros.ButtonSets.ShowTypeOutCombatMouse = 6

function XMacros.ButtonSets.Create(parentFrame, sets, roleSets, tooltip, getMacroCallback)
	local buttonSets = {}
	
	buttonSets.parentFrame = parentFrame
	buttonSets.sets = sets
	buttonSets.roleSets = roleSets
	buttonSets.tooltip = tooltip
	buttonSets.getMacroCallback = getMacroCallback
	if not buttonSets.sets then
		buttonSets.sets = {}
	end
	
	if not buttonSets.roleSets then
		buttonSets.roleSets = {}
	end
		
	return buttonSets
end

function XMacros.ButtonSets.GetSetNames(buttonSets)
	local setNames = {}
	for name, _ in pairs(buttonSets.sets) do
		table.insert(setNames, name)
	end
	
	table.sort(setNames)
	
	return setNames
end

function XMacros.ButtonSets.GetRoleSets(buttonSets)
	return buttonSets.roleSets
end

function XMacros.ButtonSets.GetRoleSet(buttonSets, role)
	return buttonSets.roleSets["Role" .. role]
end

function XMacros.ButtonSets.GetRoleSetNames(buttonSets, role)
	local setList = XMacros.ButtonSets.GetRoleSet(buttonSets, role)
	if setList then
		local setNames = ""
		for _, name in ipairs(setList) do
			if setName ~= "" then
				setNames = setNames .. ","
			end
			
			setNames = setNames .. name
		end
	
		return setNames
	end
end

function XMacros.ButtonSets.SetRoleSet(buttonSets, role, setNames)
	if setNames == nil then
		return
	end
	
	local roleNumber = tonumber(role)
	if roleNumber ~= nil and roleNumber >= 1 and roleNumber <= XMacros.MaxRoles then
		local setList = setNames
		if type(setNames) == "string" then
			setList = XMacros.Utils.Split(setNames, ",")
		end
		
		table.sort(setList)
		buttonSets.roleSets["Role" .. role] = setList
		
		for _, name in ipairs(setList) do
			XMacros.ButtonSets.AddSet(buttonSets, name)
		end
	else
		print("Cannot set button setName for role: " .. tostring(role))
	end
end

function XMacros.ButtonSets.GetSets(buttonSets)
	return buttonSets.sets
end

function XMacros.ButtonSets.AddSet(buttonSets, inSetName)
	if not inSetName then
		return false
	end
	
	local setName = XMacros.Utils.Trim(inSetName)
	if setName:len() < 1 then
		return false
	end
	
	if string.find(setName, ",") ~= nil then
		return false
	end
	
	if not buttonSets.sets[setName] then
		buttonSets.sets[setName] = { barSpecs = {} }
		return true
	end
	
	return false
end

function XMacros.ButtonSets.RenameSet(buttonSets, setName, newName)
	if buttonSets.sets[setName] then
		local setValue = buttonSets.sets[setName]
		buttonSets.sets[setName] = nil
		buttonSets.sets[newName] = setValue

		if buttonSets.bars then
			if buttonSets.bars[setName] then
				local barValue = buttonSets.bars[setName]
				buttonSets.bars[setName] = nil
				buttonSets.bars[newName] = barValue
			end
		end
		
		for role = 1, XMacros.MaxRoles do
			local setList = XMacros.ButtonSets.GetRoleSet(buttonSets, role)
			if setList then
				for row = 1, #setList do
					if setList[row] == setName then
						setList[row] = newName
						break
					end
				end
			end
		end
	end
end

function XMacros.ButtonSets.DeleteSet(buttonSets, setName)
	if buttonSets.sets[setName] then
		buttonSets.sets[setName] = nil

		if buttonSets.bars then
			if buttonSets.bars[setName] then
				for _, buttonBar in pairs(buttonSets.bars[setName]) do
					XMacros.ButtonBar.SetVisible(buttonBar, false)
				end
				buttonSets.bars[setName] = nil
			end
		end
		
		for role = 1, XMacros.MaxRoles do
			local setList = XMacros.ButtonSets.GetRoleSet(buttonSets, role)
			if setList then
				for row = 1, #setList do
					if setList[row] == setName then
						table.remove(setList, row)
						break
					end
				end
			end
		end
	end
end

function XMacros.ButtonSets.AddBar(buttonSets, setName, x, y, rows, cols, textureSize, combatShow)
	local buttonSet = buttonSets.sets[setName]
	if not buttonSet then
		XMacros.ButtonSets.AddSet(buttonSets, setName)
		buttonSet = buttonSets.sets[setName]
	end
	
	if buttonSet then
		local barName = XMacros.ButtonSets.GetDefaultBarName(buttonSets, setName)
		if not buttonSet.barSpecs[barName] then
			local newBar = {
				x = x,
				y = y,
				rows = rows,
				cols = cols,
				textureSize = textureSize,
				combatShow = combatShow,
				buttonSpecs = {}
				}
			buttonSet.barSpecs[barName] = newBar
		end
	end
end

function XMacros.ButtonSets.GetBars(buttonSets, setName)
	if not buttonSets.bars then
		buttonSets.bars = {}
	end
	
	local setBars = buttonSets.bars[setName]
	if not setBars then
		buttonSets.bars[setName] = {}
		setBars = buttonSets.bars[setName]
	end
	
	return setBars
end

function XMacros.ButtonSets.GetBar(buttonSets, setName, barName)
	local bars = XMacros.ButtonSets.GetBars(buttonSets, setName)
	return bars[barName]
end

function XMacros.ButtonSets.EnableSets(buttonSets, setNames, editCallback)
	if not setNames then
		return
	end
	
	if buttonSets.bars then	
		local setNameLookup = {}
		for _, setName in ipairs(setNames) do
			setNameLookup[setName] = 1
		end

		local setsToRemove = {}
		for setName, _ in pairs(buttonSets.bars) do
			if not setNameLookup[setName] then
				table.insert(setsToRemove, setName)
			end
		end

		for _, setName in ipairs(setsToRemove) do
			buttonSets.bars[setName] = nil
		end
	end
	
	for _, setName in ipairs(setNames) do
		local buttonSet = buttonSets.sets[setName]
		if buttonSet then
			local bars = XMacros.ButtonSets.GetBars(buttonSets, setName)

			local barsToDelete = {}			
			for barName, barSpec in pairs(buttonSet.barSpecs) do
				bars[barName] = nil
				
				if not barSpec.x or not barSpec.y or not barSpec.rows or not barSpec.cols or not barSpec.textureSize then
					print("Ignoring corrupt bar " .. tostring(barName) .. " in set " .. tostring(setName))
					table.insert(barsToDelete, barName)
				elseif barSpec.rows < 1 or barSpec.rows > 20 or barSpec.cols < 1 or barSpec.cols > 20 then
					print("Ignoring corrupt bar with bad dimensions" .. tostring(barName) .. " in set " .. tostring(setName))
					table.insert(barsToDelete, barName)
				else
					local mouseOverBar = false
					local combatShow = barSpec.combatShow
					if combatShow then
						if combatShow == XMacros.ButtonSets.ShowTypeMouse or combatShow == XMacros.ButtonSets.ShowTypeInCombatMouse or combatShow == XMacros.ButtonSets.ShowTypeOutCombatMouse then
							mouseOverBar = true
						end
					end
					
					local bar = XMacros.ButtonBar.Create(setName, barName, buttonSets.parentFrame, UIParent, barSpec, buttonSets.tooltip, mouseOverBar, editCallback)
					for buttonName, buttonSpec in pairs(barSpec.buttonSpecs) do
						local row, col = XMacros.ButtonBar.GetButtonPosition(buttonName)
						if buttonSpec.macros then
							for macroName, macro in pairs(buttonSpec.macros) do
								if not editCallback then
									local macroString = buttonSets.getMacroCallback(macro.category, macro.macro)
									XMacros.ButtonBar.AddMacro(bar, row, col, macroName, macro.category, macro.macro, macroString)
								end
							end
						end
					end

					bars[barName] = bar
				end
			end
			
			for _, barName in ipairs(barsToDelete) do
				buttonSet.barSpecs[barName] = nil
			end
		end
	end
end

function XMacros.ButtonSets.MakeSetsVisible(buttonSets, setNames, visible, inCombat)
	if setNames then
		for _, setName in ipairs(setNames) do
			local buttonSet = buttonSets.sets[setName]
			if buttonSet then
				if visible == false then
					local bars = XMacros.ButtonSets.GetBars(buttonSets, setName)
					for _, buttonBar in pairs(bars) do
						XMacros.ButtonBar.SetVisible(buttonBar, visible)
					end
				else
					for barName, barSpec in pairs(buttonSet.barSpecs) do
						local buttonBar = XMacros.ButtonSets.GetBar(buttonSets, setName, barName)
						if buttonBar then
							local visible = true
							local combatShow = barSpec.combatShow
							if combatShow then
								if combatShow == XMacros.ButtonSets.ShowTypeInCombat or combatShow == XMacros.ButtonSets.ShowTypeInCombatMouse then
									visible = inCombat
								elseif combatShow == XMacros.ButtonSets.ShowTypeOutCombat or combatShow == XMacros.ButtonSets.ShowTypeOutCombatMouse then
									visible = not inCombat
								end
							end					

							XMacros.ButtonBar.SetVisible(buttonBar, visible)
						end
					end
				end
			end
		end
	end
end

function XMacros.ButtonSets.MakeSetsDragable(buttonSets, setNames, visible)
	if setNames then
		for _, setName in ipairs(setNames) do
			local buttonSet = buttonSets.sets[setName]
			if buttonSet then
				local bars = XMacros.ButtonSets.GetBars(buttonSets, setName)
				for barName, buttonBar in pairs(bars) do
					XMacros.ButtonBar.SetDragEnabled(buttonBar, visible, function(x, y) XMacros.ButtonSets.SetBarScreenPosition(buttonSets, setName, barName, x, y) end)
					XMacros.ButtonBar.SetVisible(buttonBar, true, visible)
				end
			end
		end
	end
end

function XMacros.ButtonSets.GetBarNames(buttonSets, setName)
	local barNames = {}
	
	if buttonSets.sets[setName] then
		for name, _ in pairs(buttonSets.sets[setName]) do
			table.insert(barNames, name)
		end
	
		table.sort(barNames)
	end
	
	return barNames
end

function XMacros.ButtonSets.GetDefaultBarName(buttonSets, setName)
	if buttonSets.sets[setName] then
		local barNames = {}
		for name, _ in pairs(buttonSets.sets[setName].barSpecs) do
			barNames[name] = 1
		end
		
		for i = 1, 100 do
			local barName = "Bar" .. tostring(i)
			if not barNames[barName] then
				return barName
			end
		end
	else
		return "Bar1"
	end
end


function XMacros.ButtonSets.SetBarScreenPosition(buttonSets, setName, barName, x, y)
	local buttonSet = buttonSets.sets[setName]
	if buttonSet then
		local barSpec = buttonSet.barSpecs[barName]
		if barSpec then
			barSpec.x = x
			barSpec.y = y
		end
	end
end

function XMacros.ButtonSets.GetBarSpec(buttonSets, setName, barName)
	local buttonSet = buttonSets.sets[setName]
	if buttonSet then
		return buttonSet.barSpecs[barName]
	end
end

function XMacros.ButtonSets.GetButtonSpec(buttonSets, setName, barName, row, col)
	local barSpec = XMacros.ButtonSets.GetBarSpec(buttonSets, setName, barName)
	if barSpec then
		if not barSpec.buttonSpecs then
			barSpec.buttonSpecs = {}
		end

		local buttonName = XMacros.ButtonBar.GetButtonName(row, col)
		local button = barSpec.buttonSpecs[buttonName]
		if not button then
			barSpec.buttonSpecs[buttonName] = {}
			button = barSpec.buttonSpecs[buttonName]
		end
		
		return button
	end
end

function XMacros.ButtonSets.SetMacro(buttonSets, setName, barName, row, col, macroButton, macroCategory, macroName)
	local button = XMacros.ButtonSets.GetButtonSpec(buttonSets, setName, barName, row, col)
	if button then
		if not button.macros then
			button.macros = {}
		end

		button.macros[macroButton] = { category = macroCategory, macro = macroName }
	end
end

function XMacros.ButtonSets.SetTexture(buttonSets, setName, barName, row, col, textureAddon, texturePath)
	local button = XMacros.ButtonSets.GetButtonSpec(buttonSets, setName, barName, row, col)
	if button then
		button.textureAddon = textureAddon
		button.textureName = texturePath
	end
end

function XMacros.ButtonSets.SetAbilityName(buttonSets, setName, barName, row, col, abilityName)
	local button = XMacros.ButtonSets.GetButtonSpec(buttonSets, setName, barName, row, col)
	if button then
		button.abilityName = abilityName
	end
end

function XMacros.ButtonSets.SetBarSize(buttonSets, setName, barName, iconSize)
	local barSpec = XMacros.ButtonSets.GetBarSpec(buttonSets, setName, barName)
	if barSpec then
		barSpec.textureSize = iconSize
	end
end

function XMacros.ButtonSets.DeleteBar(buttonSets, setName, barName)
	local buttonSet = buttonSets.sets[setName]
	if buttonSet then
		buttonSet.barSpecs[barName] = nil
	end
end

